/***********************************************************************************************************************************
	Using Emergency Mode to access a RECOVERY PENDING or SUSPECT database

	Useful Links:
		https://www.sqlskills.com/blogs/paul/search-engine-qa-4-using-emergency-mode-to-access-a-recovery-pending-or-suspect-database/
		https://www.sqlskills.com/blogs/paul/corruption-last-resorts-that-people-try-first/
		https://www.sqlskills.com/blogs/paul/dbcc-writepage/
		https://www.sqlskills.com/blogs/paul/checkdb-from-every-angle-emergency-mode-repair-the-very-very-last-resort/
		https://www.sqlskills.com/blogs/paul/disaster-recovery-101-hack-attach-a-damaged-database/
		https://blogs.msdn.microsoft.com/sql/2015/04/22/how-to-attach-a-sql-server-database-without-a-transaction-log-and-with-open-transactions/

	Corruption violates the Consistency principle of the ACID protocol. Below we will see the database in an inconsistent state.
************************************************************************************************************************************/

USE master;
GO

--databases and their state
SELECT [name], database_id, user_access_desc, state_desc  FROM sys.databases;

/*
	...the state the database is in determines what you can do if you dont have a backup. 
	The easiest case is when its still ONLINE. In this case you can probably run repair to remove the damage, most likely with some data loss, 
	and then take steps to prevent the damage occurring again. 
	If repair cant fix all the errors then your only option without a backup is to extract as much data as you can into a new database.
*/

--If the DB is in RECOVERY or SUSPECT states then we need to get into the DB to learn whether it is just logically inconcistent or physically inconsistent


/*******************************************************************
	Setup example to test
********************************************************************/

CREATE DATABASE [EmergencyDemo];
GO

USE [EmergencyDemo];
GO
 
CREATE TABLE [salaries] (
    [FirstName]    CHAR (20),
    [LastName]     CHAR (20),
    [Salary]       INT
);
GO
 
INSERT INTO [salaries] (FirstName, LastName, Salary) VALUES ('Vincent', 'Vega', 100000);
INSERT INTO [salaries] (FirstName, LastName, Salary) VALUES ('Jules', 'Winnfield', 120000);
INSERT INTO [salaries] (FirstName, LastName, Salary) VALUES ('Mia', 'Wallace', 110000);
GO

SELECT * FROM dbo.salaries; --take note of the output


/*******************************************************************
	Corrupt the database
********************************************************************/

--starting an explicit transaction to update a salary for someone
BEGIN TRANSACTION;
 	UPDATE [salaries] 
	SET [Salary] = 0 
	WHERE LastName = 'Vega';

--force the data page holding the updated row to be flushed to disk (written to the xact log)
CHECKPOINT;

--REMEMBER: this is still an UNCOMMITTED transaction
--We have an active, uncommitted transaction thats modified the table, and the table modification has been written to disk. 
--If the power failed at this point, crash recovery would run and the transaction would be rolled back. 
--We can simulate this by shutting down SQL Server. 

--In another connection (OR SSMS OR SQL Configuration Manager OR services):
SHUTDOWN WITH NOWAIT;

--To make matters worse let's simulate damage to the xact log by deleting it
--Start SQL Server services back up
--Check SQL Error Log...also notice dump files
--Refresh Object Explorer for databases


/*******************************************************************
	Try to gain access
********************************************************************/

--Try to access the database
USE EmergencyDemo; --knock, knock...
GO

--It is time to break out EMERGENCY mode
--requires sysadmin
--the issue with EMERGENCY mode: recovery hasn't run yet so we do not know if the db is logically or phsically consistent. 
--need to get in db to look around and possibly fix
ALTER DATABASE EmergencyDemo SET EMERGENCY;
GO

--Try to access the database
USE EmergencyDemo;
GO

SELECT * FROM dbo.salaries; --inconsistent output

DBCC CHECKDB('EmergencyDemo') WITH NO_INFOMSGS; --nope

--Try to bring the database online again
ALTER DATABASE EmergencyDemo SET ONLINE; --nope
--Refresh Object Explorer for databases and note database status


/*******************************************************************
	How to fix? Try various ideas...
********************************************************************/

--Maybe adding transaction log file will help?
ALTER DATABASE EmergencyDemo
ADD LOG FILE (
	 NAME = EmergencyDemo_log_new
	,FILENAME = 'C:\Program Files\Microsoft SQL Server\MSSQL13.QUASAR\MSSQL\DATA\EmergencyDemo.ldf'
	,SIZE = 5MB
	,MAXSIZE = 100MB
	,FILEGROWTH = 5MB
); --how did that go?

--OK I will detach the database then reattach it. Maybe SQL Server will create me a new xact log?
EXEC sp_detach_db EmergencyDemo, 'true';

--Try to attach 
USE [master]
GO
CREATE DATABASE [EmergencyDemo] ON 
( FILENAME = N'C:\Program Files\Microsoft SQL Server\MSSQL13.QUASAR\MSSQL\DATA\EmergencyDemo.mdf' )
FOR ATTACH
GO --how did that go?


/*******************************************************************
	Desperate times call for...
	
	How to attach a db without a transaction log and with open transactions
********************************************************************/

USE master;
GO

DBCC TRACEON(3604); --redirect output to client
GO

ALTER DATABASE EmergencyDemo SET EMERGENCY;
GO

ALTER DATABASE EmergencyDemo SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
GO    

DBCC CHECKDB('EmergencyDemo', REPAIR_ALLOW_DATA_LOSS) WITH NO_INFOMSGS;
GO

ALTER DATABASE EmergencyDemo SET MULTI_USER;
GO

------------------------------------------------------------------------------------------

--check output again
SELECT * FROM EmergencyDemo.dbo.salaries;


--Cleanup
DROP DATABASE EmergencyDemo;
